// $Id: Query.java,v 1.2 2007-08-24 12:39:51 pierre Exp $

package com.senellart.pierre.fuzzyxml.query;

import java.io.StringWriter;
import java.util.Iterator;
import java.util.Set;

import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.stream.StreamResult;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.traversal.DocumentTraversal;
import org.w3c.dom.traversal.NodeFilter;
import org.w3c.dom.traversal.NodeIterator;

import com.senellart.pierre.fuzzyxml.FuzzyXML;
import com.senellart.pierre.fuzzyxml.document.DocumentManager;
import com.senellart.pierre.fuzzyxml.document.FXMLDocument;
import com.senellart.pierre.fuzzyxml.event.EventConjunction;
import com.senellart.pierre.fuzzyxml.exception.InvalidQueryException;
import com.senellart.pierre.fuzzyxml.exception.QueryProcessingException;

/**
 * @author pierre
 * 
 */
public abstract class Query {
	protected DocumentManager documentManager;

	String queryString;

	/**
	 * @param environment
	 * @param transform
	 * @param connection
	 * @param query
	 */
	public Query(DocumentManager documentManager, Source query)
			throws InvalidQueryException {
		this.documentManager = documentManager;
		StringWriter queryWriter = new StringWriter();

		try {
			Transformer transform = documentManager.getQueryFactory()
					.getQuery2XQueryStylesheet();
			transform.transform(query, new StreamResult(queryWriter));
			queryString = queryWriter.toString();
			System.err.println(queryString);
		} catch (Exception e) {
			throw new InvalidQueryException(e);
		}
	}

	public Document execute(FXMLDocument fxmlDoc)
			throws QueryProcessingException {
		Document doc = xQueryExecute(fxmlDoc);
		moveEvents(doc);
		filterDropElements(doc);
		return doc;
	}

	abstract public Document xQueryExecute(FXMLDocument fxmlDoc)
			throws QueryProcessingException;

	/**
	 * @author pierre
	 * 
	 */
	private static class FXMLConditionFilter implements NodeFilter {
		/*
		 * (non-Javadoc)
		 * 
		 * @see org.w3c.dom.traversal.NodeFilter#acceptNode(org.w3c.dom.Node)
		 */
		public short acceptNode(Node n) {
			if (n.getNodeType() == Node.ELEMENT_NODE) {
				if (n.getLocalName().equals("event")
						&& FuzzyXML.NAMESPACE_URI.equals(n.getNamespaceURI()))
					return FILTER_ACCEPT;
			}

			return FILTER_SKIP;
		}
	}

	public static boolean getConditions(Node root, EventConjunction events, boolean remove) {
		NodeIterator it = ((DocumentTraversal) root.getOwnerDocument())
				.createNodeIterator(root, NodeFilter.SHOW_ALL,
						new FXMLConditionFilter(), true);

		Node next = it.nextNode();
		while (next != null) {
			Node n = next;
			next = it.nextNode();

			if (n.getNodeType() != Node.ELEMENT_NODE)
				continue;

			Element el = (Element) n;

			String name = el.getAttribute("id");

			boolean pos = !el.hasAttribute("neg");
			if (events.contains(name)) {
				if (events.get(name) != pos)
					return true;
			} else {
				events.add(name, pos);
			}

			if (remove) {
				Node parent = n.getParentNode(); // fxml:condition node
				Node greatParent = parent.getParentNode();
				if (greatParent != null) // May already be removed
					greatParent.removeChild(parent);
			}
		}
		
		return false;
	}

	/**
	 * @param doc
	 */
	protected static void moveEvents(Document doc) {
		Node root = doc.getDocumentElement();
		Node resultNext = null;
		for (Node result = root.getFirstChild(); result != null; result = resultNext) {
			resultNext = result.getNextSibling();

			EventConjunction events = new EventConjunction();
			boolean inconsistent = getConditions(result, events, true);

			if (inconsistent) {
				root.removeChild(result);
			} else if (!events.isTautology()) {
				Element condition = doc.createElementNS(FuzzyXML.NAMESPACE_URI,
						"fxml:condition");

				Set eventNames = events.getIds();

				for (Iterator eventIt = eventNames.iterator(); eventIt
						.hasNext();) {
					String name = (String) eventIt.next();

					Element event = doc.createElementNS(FuzzyXML.NAMESPACE_URI,
							"fxml:event");
					event.setAttribute("id", name);
					if (!events.get(name))
						event.setAttribute("neg", "1");

					condition.appendChild(event);
				}

				result.insertBefore(condition, result.getFirstChild());
			}
		}
	}

	/**
	 * @author pierre
	 * 
	 */
	private static class FXMLDropFilter implements NodeFilter {
		/*
		 * (non-Javadoc)
		 * 
		 * @see org.w3c.dom.traversal.NodeFilter#acceptNode(org.w3c.dom.Node)
		 */
		public short acceptNode(Node n) {
			if (n.getNodeType() == Node.ELEMENT_NODE
					&& !FuzzyXML.NAMESPACE_URI.equals(n.getNamespaceURI())) {
				if (((Element) n).hasAttributeNS(FuzzyXML.NAMESPACE_URI,
						"nodrop")) {
					return FILTER_SKIP;
				} else
					return FILTER_ACCEPT;
			}

			return FILTER_SKIP;
		}

	}

	/**
	 * @param doc
	 */
	protected static void filterDropElements(Document doc) {
		Node root = doc.getDocumentElement();

		NodeIterator it = ((DocumentTraversal) doc).createNodeIterator(root,
				NodeFilter.SHOW_ALL, new FXMLDropFilter(), true);

		Node next = it.nextNode();
		while (next != null) {
			Node n = next;
			next = it.nextNode();

			// Before dropping it, check whether a descendant has a fxml:nodrop
			if (hasNoDrop(n.getChildNodes()))
				continue;

			Node parent = n.getParentNode();
			if (parent != null)
				parent.removeChild(n);
		}
	}

	/**
	 * @param nl
	 * @return
	 */
	private static boolean hasNoDrop(NodeList nl) {
		if (nl == null)
			return false;

		int size = nl.getLength();
		for (int i = 0; i < size; ++i) {
			Node n = nl.item(i);
			if (n.getNodeType() == Node.ELEMENT_NODE) {
				if (((Element) n).hasAttributeNS(FuzzyXML.NAMESPACE_URI,
						"nodrop"))
					return true;
				if (hasNoDrop(n.getChildNodes()))
					return true;
			}
		}

		return false;
	}
}
