// $Id: Update.java,v 1.2 2007-08-24 12:39:52 pierre Exp $

package com.senellart.pierre.fuzzyxml.update;

import java.io.InputStream;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.SortedMap;
import java.util.TreeMap;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.traversal.DocumentTraversal;
import org.w3c.dom.traversal.NodeFilter;
import org.w3c.dom.traversal.NodeIterator;

import com.senellart.pierre.fuzzyxml.FuzzyXML;
import com.senellart.pierre.fuzzyxml.document.DocumentManager;
import com.senellart.pierre.fuzzyxml.document.FXMLDocument;
import com.senellart.pierre.fuzzyxml.event.EventConjunction;
import com.senellart.pierre.fuzzyxml.event.EventConjunctiveFormula;
import com.senellart.pierre.fuzzyxml.event.EventDisjunctiveFormula;
import com.senellart.pierre.fuzzyxml.event.EventNegativeFormula;
import com.senellart.pierre.fuzzyxml.exception.InvalidQueryException;
import com.senellart.pierre.fuzzyxml.exception.InvalidUpdateException;
import com.senellart.pierre.fuzzyxml.exception.UpdateProcessingException;
import com.senellart.pierre.fuzzyxml.query.Query;

/**
 * @author pierre
 * 
 */
public abstract class Update {
	private class FXMLUpdateFilter implements NodeFilter {

		public short acceptNode(Node n) {
			if (n.getNodeType() == Node.ELEMENT_NODE
					&& FuzzyXML.NAMESPACE_URI.equals(n.getNamespaceURI())
					&& (n.getLocalName().equals("delete") || n.getLocalName()
							.equals("insert")))
				return FILTER_ACCEPT;
			else
				return FILTER_SKIP;
		}
	}

	protected DocumentManager documentManager;

	protected Query query;

	private String newEventId;

	/**
	 * @param environment
	 * @param queryFactory
	 * @param source
	 */
	protected Update(DocumentManager documentManager, InputStream source)
			throws InvalidUpdateException {
		this.documentManager = documentManager;
		newEventId = null;
		try {
			query = documentManager.getQueryFactory().newQuery(source);
		} catch (InvalidQueryException e) {
			throw new InvalidUpdateException(e);
		}
	}
	
	protected Update(DocumentManager documentManager, Document source)
			throws InvalidUpdateException {
		this.documentManager = documentManager;
		newEventId = null;
		try {
			query = documentManager.getQueryFactory().newQuery(source);
		} catch (InvalidQueryException e) {
			throw new InvalidUpdateException(e);
		}
	}

	public void execute(FXMLDocument fxmlDoc) throws UpdateProcessingException {
		Document queryResult;

		try {
			queryResult = query.xQueryExecute(fxmlDoc);
			Document xUpdateOperation = buildXUpdate(queryResult, fxmlDoc);

			if (xUpdateOperation != null) {
				applyUpdateOperation(xUpdateOperation, fxmlDoc);
			}
		} catch (Exception e) {
			throw new UpdateProcessingException(e);
		}
	}

	abstract protected void applyUpdateOperation(Document updateOperation,
			FXMLDocument fxmlDoc) throws UpdateProcessingException;

	private static class OperationInfo {
		private EventConjunction events;

		private Node node;

		public EventConjunction getEvents() {
			return events;
		}

		public Node getNode() {
			return node;
		}

		OperationInfo(EventConjunction newEvents, Node node) {
			this.events = newEvents;
			this.node = node;
		}
	};

	private static class PrefixComparator implements Comparator {
		// Total order on strings, extension of the prefix order (which
		// is partial)
		public int compare(Object arg0, Object arg1) {
			String a = (String) arg0;
			String b = (String) arg1;

			if (a.length() < b.length())
				return -1;
			else if(a.length() > b.length())
				return 1;
			return (a.compareTo(b));
		}
	}

	protected Document buildXUpdate(Document doc, FXMLDocument fxmlDoc) {
		Document xUpdate = documentManager.getEnvironment().getDOMParser()
				.getDOMImplementation().createDocument(
						FuzzyXML.XUPDATE_NAMESPACE_URI, "modifications", null);
		Node root = doc.getDocumentElement();
		double confidence = Double.parseDouble(((Element) root)
				.getAttribute("confidence"));
		String origin = ((Element) root).getAttribute("origin");

		SortedMap operationsByLocator = new TreeMap(new PrefixComparator());

		for (Node result = root.getFirstChild(); result != null; result = result
				.getNextSibling()) {
			if (result.getNodeType() != Node.ELEMENT_NODE)
				continue;

			EventConjunction events = new EventConjunction();

			boolean inconsistent = Query.getConditions(result, events, false);
			
			if (!inconsistent) {
				if (newEventId == null)
					newEventId = fxmlDoc.addEvent(confidence, origin);
				events.add(newEventId,true);

				NodeIterator it = ((DocumentTraversal) doc).createNodeIterator(
						result, NodeFilter.SHOW_ALL, new FXMLUpdateFilter(),
						true);

				Node operation;
				for (operation = it.nextNode(); operation != null; operation = it
						.nextNode()) {
					String xPathLocator = getXPathLocator(operation
							.getParentNode());

					if (!operationsByLocator.containsKey(xPathLocator))
						operationsByLocator.put(xPathLocator, new HashMap());

					String nodeName = operation.getLocalName();

					Map map = (Map) operationsByLocator.get(xPathLocator);
					if (!map.containsKey(nodeName))
						map.put(nodeName, new HashSet());
					((Set) map.get(nodeName)).add(new OperationInfo(events,
							operation));
				}
			}
		}

		buildXUpdateLocal(fxmlDoc, xUpdate, null, operationsByLocator);

		if (xUpdate.getDocumentElement().getChildNodes().getLength() == 0)
			return null;
		return xUpdate;
	}

	private void buildXUpdateLocal(FXMLDocument fxmlDoc, Document xUpdate,
			Node operationPosition, SortedMap operationsByLocator) {
		Set locatorSet = operationsByLocator.keySet();
		Set locatorsAlreadyDone = new HashSet();
		for (Iterator locatorIt = locatorSet.iterator(); locatorIt.hasNext();) {
			String locator = (String) locatorIt.next();

			if (locatorsAlreadyDone.contains(locator))
				continue;

			Map operations = (Map) operationsByLocator.get(locator);

			OperationInfo someInfo = ((OperationInfo) ((Set) operations
					.get(operations.keySet().iterator().next())).iterator()
					.next());

			EventConjunction ancestorEvents = new EventConjunction();
			EventConjunction localEvents = new EventConjunction();
			getAncestorEvents(someInfo.getNode().getParentNode(),
					ancestorEvents);
			getEvents(someInfo.getNode().getParentNode(), localEvents);

			if (!operations.containsKey("delete")) {
				// Great, no particular difficulty :-)
				// Just insert every node with their conditions

				Set insertionsSet = ((HashSet) operations.get("insert"));
				for (Iterator operationsIt = insertionsSet.iterator(); operationsIt
						.hasNext();) {
					OperationInfo info = (OperationInfo) operationsIt.next();

					Node toInsert = info.getNode().getChildNodes().item(0);
					while (toInsert.getNodeType() != Node.ELEMENT_NODE)
						toInsert = toInsert.getNextSibling();

					EventConjunction newEvents = info.getEvents().except(
							ancestorEvents);

					performInsert(xUpdate, operationPosition, locator,
							toInsert, newEvents, fxmlDoc);
				}
			} else {
				// Delete the node
				performDelete(xUpdate, operationPosition, locator);

				// Gather the conditions of all deletions
				Set deleteSet = ((Set) operations.get("delete"));
				EventDisjunctiveFormula F = new EventDisjunctiveFormula();
				for (Iterator operationsIt = deleteSet.iterator(); operationsIt
						.hasNext();) {
					EventConjunction newEvents = ((OperationInfo) operationsIt
							.next()).getEvents().except(ancestorEvents);

					F.addTerm(newEvents);
				}
				EventConjunctiveFormula G = new EventConjunctiveFormula();
				G.addTerm(localEvents);
				G.addTerm(new EventNegativeFormula(F));
				EventDisjunctiveFormula H = G.getDNF();

				// Get all opertions whose locator is a prefix of this one
				SortedMap childrenOperation = new TreeMap(
						new PrefixComparator());
				Iterator locatorIt2 = locatorSet.iterator();
				{
					String locator2;
					do {
						locator2 = (String) locatorIt2.next();
					} while (!locator2.equals(locator));
				}
				while (locatorIt2.hasNext()) {
					String locator2 = (String) locatorIt2.next();
					if (locator2.substring(0, locator.length()).equals(locator)) {
						childrenOperation.put(locator2.substring(locator
								.length() + 1), operationsByLocator
								.get(locator2));
						locatorsAlreadyDone.add(locator2);
					}
				}

				String parentLocator = locator.substring(0, locator
						.lastIndexOf('/'));

				// Add as many nodes as elements of H
				for (Iterator it = H.getTerms().iterator(); it.hasNext();) {
					Node toInsert = someInfo.getNode().getParentNode()
							.cloneNode(true);
					buildXUpdateLocal(fxmlDoc, null, toInsert,
							childrenOperation);

					EventConjunction term = (EventConjunction) it.next();

					performInsert(xUpdate, operationPosition, parentLocator,
							toInsert, term, fxmlDoc);
				}
			}
		}
	}

	private void performDelete(Document xUpdate, Node operationPosition,
			String locator) {
		if (xUpdate != null)
			performDeleteXUpdate(xUpdate, locator);
		else
			performDeleteLocal(operationPosition, locator);
	}

	private void performDeleteXUpdate(Document xUpdate, String locator) {
		xUpdate.getDocumentElement().appendChild(
				createXUpdateOperation("delete", xUpdate, locator));
	}

	private void performDeleteLocal(Node operationPosition, String locator) {
		// TODO Auto-generated method stub
System.err.println("not implemented");
	}

	private void performInsert(Document xUpdate, Node operationPosition,
			String locator, Node toInsert, EventConjunction events,
			FXMLDocument fxmlDoc) {
		if (xUpdate != null)
			performInsertXUpdate(xUpdate, locator, toInsert, events, fxmlDoc);
		else
			performInsertLocal(operationPosition, locator, toInsert, events,
					fxmlDoc);
	}

	private void performInsertXUpdate(Document xUpdate, String locator,
			Node toInsert, EventConjunction events, FXMLDocument fxmlDoc) {
		Element xUpdateOperation = createXUpdateOperation("append", xUpdate,
				locator);

		Element xUpdateElement = createXUpdateElement(xUpdate, toInsert,
				fxmlDoc, events);

		xUpdateOperation.appendChild(xUpdateElement);
		xUpdate.getDocumentElement().appendChild(xUpdateOperation);
	}

	private void performInsertLocal(Node operationPosition, String locator,
			Node toInsert, EventConjunction events, FXMLDocument fxmlDoc) {
		// TODO Auto-generated method stub
		System.err.println("not implemented");
	}

	private Element createXUpdateElement(Document xUpdate, Node toInsert,
			FXMLDocument fxmlDoc, EventConjunction newEvents) {
		Element xUpdateElement = xUpdate.createElementNS(
				FuzzyXML.XUPDATE_NAMESPACE_URI, "element");

		xUpdateElement.setAttribute("name", toInsert.getLocalName());
		xUpdateElement.setAttribute("namespace", toInsert.getNamespaceURI());

		Element xUpdateAttribute = xUpdate.createElementNS(
				FuzzyXML.XUPDATE_NAMESPACE_URI, "attribute");
		xUpdateAttribute.setAttribute("name", "fxml:id");
		xUpdateAttribute.setAttribute("namespace", FuzzyXML.NAMESPACE_URI);
		xUpdateAttribute.appendChild(xUpdate.createTextNode(fxmlDoc.getNewId()));
		xUpdateElement.appendChild(xUpdateAttribute);

		Element condition = xUpdate.createElementNS(FuzzyXML.NAMESPACE_URI,
				"fxml:condition");
		xUpdateElement.appendChild(condition);

		Set newEventSet = newEvents.getIds();
		for (Iterator eventIt = newEventSet.iterator(); eventIt.hasNext();) {
			String e = (String) eventIt.next();
			Element event = xUpdate.createElementNS(FuzzyXML.NAMESPACE_URI,
					"fxml:event");

			if (!newEvents.get(e))
				event.setAttribute("neg", "1");

			event.setAttribute("id", e);

			condition.appendChild(event);
		}

		NodeList toInsertChildren = toInsert.getChildNodes();
		int toInsertChildrenSize = toInsertChildren.getLength();
		for (int i = 0; i < toInsertChildrenSize; ++i) {
			Node localNode = toInsertChildren.item(i);
			if (localNode.getNodeType() == Node.ELEMENT_NODE
					&& FuzzyXML.NAMESPACE_URI.equals(localNode
							.getNamespaceURI())
					&& !"text".equals(localNode.getLocalName()))
				continue;
			Node n = xUpdate.importNode(localNode,true);
			if (n.getNodeType() == Node.TEXT_NODE) {
				Element fxmlText = xUpdate.createElementNS(
						FuzzyXML.NAMESPACE_URI, "fxml:text");
				fxmlText.appendChild(n);
				xUpdateElement.appendChild(fxmlText);
			} else {
				xUpdateElement.appendChild(n);
			}
		}

		return xUpdateElement;
	}

	private Element createXUpdateOperation(String type, Document xUpdate,
			String xPathLocator) {
		Element xUpdateOperation = xUpdate.createElementNS(
				FuzzyXML.XUPDATE_NAMESPACE_URI, type);
		xUpdateOperation.setAttribute("select", xPathLocator);
		return xUpdateOperation;
	}

	private void getAncestorEvents(Node node, EventConjunction events) {
		if (node == null || node.getNodeType() != Node.ELEMENT_NODE)
			return;

		Update.getEvents(node, events);

		getAncestorEvents(node.getParentNode(), events);
	}

	private static void getEvents(Node node, EventConjunction events) {
		boolean ok = false;
		Node condition = node.getChildNodes().item(0);
		for (condition = node.getChildNodes().item(0); condition != null; condition = condition
				.getNextSibling()) {
			if (condition.getNodeType() != Node.ELEMENT_NODE)
				continue;
			if (!FuzzyXML.NAMESPACE_URI.equals(condition.getNamespaceURI()))
				break;
			if ("condition".equals(condition.getLocalName())) {
				ok = true;
				break;
			}
		}

		if (ok) {
			NodeList eventNodes = condition.getChildNodes();
			int size = eventNodes.getLength();
			for (int i = 0; i < size; ++i) {
				Node eventNode = eventNodes.item(i);

				if (eventNode.getNodeType() != Node.ELEMENT_NODE)
					continue;

				events.add(((Element) eventNode).getAttribute("id"),
						!((Element) eventNode).hasAttribute("neg"));
			}
		}
	}

	private static String getXPathLocator(Node node) {
		if (node == null || node.getNodeType() != Node.ELEMENT_NODE
				|| "result".equals(node.getLocalName())
				&& FuzzyXML.NAMESPACE_URI.equals(node.getNamespaceURI()))
			return "";
		else {
			String locator = node.getNodeName()
					+ "["
					+ "@fxml:id='"
					+ ((Element) node).getAttributeNS(FuzzyXML.NAMESPACE_URI,
							"id") + "']";

			return getXPathLocator(node.getParentNode()) + "/" + locator;
		}
	}
}
